@@ -13,6 +13,7 @@ import logging
 import fnmatch
 import subprocess
 import functools
+import re
 
 import apt
 
@@ -20,7 +21,8 @@ from UbuntuDrivers import kerneldetection
 
 system_architecture = apt.apt_pkg.get_architectures()[0]
 
-def system_modaliases():
+
+def system_modaliases(sys_path=None):
     '''Get modaliases present in the system.
 
     This ignores devices whose drivers are statically built into the kernel, as
@@ -29,7 +31,8 @@ def system_modaliases():
     Return a modalias → sysfs path map.
     '''
     aliases = {}
-    for path, dirs, files in os.walk('/sys/devices'):
+    devices = sys_path and '%s/devices' % (sys_path) or '/sys/devices'
+    for path, dirs, files in os.walk(devices):
         modalias = None
 
         # most devices have modalias files
@@ -39,13 +42,12 @@ def system_modaliases():
                     modalias = f.read().strip()
             except IOError as e:
                 logging.debug('system_modaliases(): Cannot read %s/modalias: %s',
-                        path, e)
+                              path, e)
                 continue
 
         # devices on SSB bus only mention the modalias in the uevent file (as
         # of 2.6.24)
         elif 'ssb' in path and 'uevent' in files:
-            info = {}
             with open(os.path.join(path, 'uevent')) as f:
                 for l in f:
                     if l.startswith('MODALIAS='):
@@ -56,16 +58,17 @@ def system_modaliases():
             continue
 
         # ignore drivers which are statically built into the kernel
-        driverlink =  os.path.join(path, 'driver')
+        driverlink = os.path.join(path, 'driver')
         modlink = os.path.join(driverlink, 'module')
         if os.path.islink(driverlink) and not os.path.islink(modlink):
-            #logging.debug('system_modaliases(): ignoring device %s which has no module (built into kernel)', path)
+            # logging.debug('system_modaliases(): ignoring device %s which has no module (built into kernel)', path)
             continue
 
         aliases[modalias] = path
 
     return aliases
 
+
 def _check_video_abi_compat(apt_cache, record):
     xorg_video_abi = None
 
@@ -74,7 +77,7 @@ def _check_video_abi_compat(apt_cache, record):
         for p in apt_cache['xserver-xorg-core'].candidate.provides:
             if p.startswith('xorg-video-abi-'):
                 xorg_video_abi = p
-                #logging.debug('_check_video_abi_compat(): Current X.org video abi: %s', xorg_video_abi)
+                # logging.debug('_check_video_abi_compat(): Current X.org video abi: %s', xorg_video_abi)
                 break
     except (AttributeError, KeyError):
         logging.debug('_check_video_abi_compat(): xserver-xorg-core not available, cannot check ABI')
@@ -87,11 +90,12 @@ def _check_video_abi_compat(apt_cache, record):
     except KeyError:
         return True
     if 'xorg-video-abi-' in deps and xorg_video_abi not in deps:
-        logging.debug('Driver package %s is incompatible with current X.org server ABI %s', 
-                record['Package'], xorg_video_abi)
+        logging.debug('Driver package %s is incompatible with current X.org server ABI %s',
+                      record['Package'], xorg_video_abi)
         return False
     return True
 
+
 def _apt_cache_modalias_map(apt_cache):
     '''Build a modalias map from an apt.Cache object.
 
@@ -106,7 +110,7 @@ def _apt_cache_modalias_map(apt_cache):
         # skip foreign architectures, we usually only want native
         # driver packages
         if (not package.candidate or
-            package.candidate.architecture not in ('all', system_architecture)):
+                package.candidate.architecture not in ('all', system_architecture)):
             continue
 
         # skip packages without a modalias field
@@ -135,9 +139,10 @@ def _apt_cache_modalias_map(apt_cache):
 
     return result
 
+
 def packages_for_modalias(apt_cache, modalias):
     '''Search packages which match the given modalias.
-    
+
     Return a list of apt.Package objects.
     '''
     pkgs = set()
@@ -157,18 +162,21 @@ def packages_for_modalias(apt_cache, modalias):
 
     return [apt_cache[p] for p in pkgs]
 
+
 packages_for_modalias.cache_maps = {}
 
+
 def _is_package_free(pkg):
     assert pkg.candidate is not None
     # it would be better to check the actual license, as we do not have
     # the component for third-party packages; but this is the best we can do
     # at the moment
-    if pkg.candidate.section.startswith('restricted') or \
-            pkg.candidate.section.startswith('multiverse'):
-        return False
+    for o in pkg.candidate.origins:
+        if o.component in ('restricted', 'multiverse'):
+            return False
     return True
 
+
 def _is_package_from_distro(pkg):
     if pkg.candidate is None:
         return False
@@ -178,6 +186,7 @@ def _is_package_from_distro(pkg):
             return True
     return False
 
+
 def _pkg_get_module(pkg):
     '''Determine module name from apt Package object'''
 
@@ -195,6 +204,7 @@ def _pkg_get_module(pkg):
     module = m[:paren]
     return module
 
+
 def _is_manual_install(pkg):
     '''Determine if the kernel module from an apt.Package is manually installed.'''
 
@@ -224,6 +234,7 @@ def _is_manual_install(pkg):
                   pkg.name, module)
     return False
 
+
 def _get_db_name(syspath, alias):
     '''Return (vendor, model) names for given device.
 
@@ -251,9 +262,10 @@ def _get_db_name(syspath, alias):
                   alias, vendor, model)
     return (vendor, model)
 
-def system_driver_packages(apt_cache=None):
+
+def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False):
     '''Get driver packages that are available for the system.
-    
+
     This calls system_modaliases() to determine the system's hardware and then
     queries apt about which packages provide drivers for those. It also adds
     available packages from detect_plugin_packages().
@@ -262,6 +274,9 @@ def system_driver_packages(apt_cache=None):
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
+    If freeonly is set to True, only free packages (from main and universe) are
+    considered
+
     Return a dictionary which maps package names to information about them:
 
       driver_package → {'modalias': 'pci:...', ...}
@@ -284,7 +299,7 @@ def system_driver_packages(apt_cache=None):
                      versions; these have this flag, where exactly one has
                      recommended == True, and all others False.
     '''
-    modaliases = system_modaliases()
+    modaliases = system_modaliases(sys_path)
 
     if not apt_cache:
         apt_cache = apt.Cache()
@@ -292,6 +307,8 @@ def system_driver_packages(apt_cache=None):
     packages = {}
     for alias, syspath in modaliases.items():
         for p in packages_for_modalias(apt_cache, alias):
+            if freeonly and not _is_package_free(p):
+                continue
             packages[p.name] = {
                     'modalias': alias,
                     'syspath': syspath,
@@ -332,9 +349,123 @@ def system_driver_packages(apt_cache=None):
 
     return packages
 
-def system_device_drivers(apt_cache=None):
+
+def _get_vendor_model_from_alias(alias):
+    modalias_pattern = re.compile('(.+):v(.+)d(.+)sv(.+)sd(.+)bc(.+)i.*')
+
+    details = modalias_pattern.match(alias)
+
+    if details:
+        return (details.group(2)[4:], details.group(3)[4:])
+
+    return (None, None)
+
+
+def _get_headless_no_dkms_metapackage(pkg, apt_cache):
+    assert pkg.candidate is not None
+    metapackage = None
+    '''Return headless-no-dkms metapackage from the main metapackage.
+
+    This is useful when dealing with packages such as nvidia-driver-$flavour
+    whose headless-no-dkms metapackage would be nvidia-headless-no-dkms-$flavour
+    '''
+    headless_template = 'nvidia-headless-no-dkms-%s'
+    name = pkg.shortname
+    flavour = name[name.rfind('-')+1:]
+    try:
+        int(flavour)
+    except ValueError:
+        return metapackage
+
+    candidate = headless_template % flavour
+
+    try:
+        package = apt_cache.__getitem__(candidate)
+        # skip foreign architectures, we usually only want native
+        # driver packages
+        if (package.candidate and
+                package.candidate.architecture in ('all', system_architecture)):
+            metapackage = candidate
+    except KeyError:
+        pass
+
+    return metapackage
+
+
+def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
+    '''Get driver packages, for gpgpu purposes, that are available for the system.
+
+    This calls system_modaliases() to determine the system's hardware and then
+    queries apt about which packages provide drivers for those. Finally, it looks
+    for the correct metapackage, by calling _get_headless_no_dkms_metapackage().
+
+    If you already have an apt.Cache() object, you should pass it as an
+    argument for efficiency. If not given, this function creates a temporary
+    one by itself.
+
+    Return a dictionary which maps package names to information about them:
+
+      driver_package → {'modalias': 'pci:...', ...}
+
+    Available information keys are:
+      'modalias':    Modalias for the device that needs this driver (not for
+                     drivers from detect plugins)
+      'syspath':     sysfs directory for the device that needs this driver
+                     (not for drivers from detect plugins)
+      'plugin':      Name of plugin that detected this package (only for
+                     drivers from detect plugins)
+      'free':        Boolean flag whether driver is free, i. e. in the "main"
+                     or "universe" component.
+      'from_distro': Boolean flag whether the driver is shipped by the distro;
+                     if not, it comes from a (potentially less tested/trusted)
+                     third party source.
+      'vendor':      Human readable vendor name, if available.
+      'model':       Human readable product name, if available.
+      'recommended': Some drivers (nvidia, fglrx) come in multiple variants and
+                     versions; these have this flag, where exactly one has
+                     recommended == True, and all others False.
+    '''
+    vendors_whitelist = ['10de']
+    modaliases = system_modaliases(sys_path)
+
+    if not apt_cache:
+        apt_cache = apt.Cache()
+
+    packages = {}
+    for alias, syspath in modaliases.items():
+        for p in packages_for_modalias(apt_cache, alias):
+            (vendor, model) = _get_db_name(syspath, alias)
+            vendor_id, model_id = _get_vendor_model_from_alias(alias)
+            if (vendor_id is not None) and (vendor_id.lower() in vendors_whitelist):
+                packages[p.name] = {
+                        'modalias': alias,
+                        'syspath': syspath,
+                        'free': _is_package_free(p),
+                        'from_distro': _is_package_from_distro(p),
+                    }
+                if vendor is not None:
+                    packages[p.name]['vendor'] = vendor
+                if model is not None:
+                    packages[p.name]['model'] = model
+                metapackage = _get_headless_no_dkms_metapackage(p, apt_cache)
+
+                if metapackage is not None:
+                    packages[p.name]['metapackage'] = metapackage
+
+    # Add "recommended" flags for NVidia alternatives
+    nvidia_packages = [p for p in packages if p.startswith('nvidia-')]
+    if nvidia_packages:
+        nvidia_packages.sort(key=functools.cmp_to_key(_cmp_gfx_alternatives))
+        recommended = nvidia_packages[-1]
+        for p in nvidia_packages:
+            packages[p]['recommended'] = (p == recommended)
+
+    return packages
+
+
+def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     '''Get by-device driver packages that are available for the system.
-    
+
     This calls system_modaliases() to determine the system's hardware and then
     queries apt about which packages provide drivers for each of those. It also
     adds available packages from detect_plugin_packages(), using the name of
@@ -344,9 +475,12 @@ def system_device_drivers(apt_cache=None):
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
+    If freeonly is set to True, only free packages (from main and universe) are
+    considered
+
     Return a dictionary which maps devices to available drivers:
 
-      device_name →  {'modalias': 'pci:...', <device info>, 
+      device_name →  {'modalias': 'pci:...', <device info>,
                       'drivers': {'pkgname': {<driver package info>}}
 
     A key (device name) is either the sysfs path (for drivers detected through
@@ -387,7 +521,8 @@ def system_device_drivers(apt_cache=None):
         apt_cache = apt.Cache()
 
     # copy the system_driver_packages() structure into the by-device structure
-    for pkg, pkginfo in system_driver_packages(apt_cache).items():
+    for pkg, pkginfo in system_driver_packages(apt_cache, sys_path,
+                                               freeonly=freeonly).items():
         if 'syspath' in pkginfo:
             device_name = pkginfo['syspath']
         else:
@@ -409,12 +544,13 @@ def system_device_drivers(apt_cache=None):
                 break
         else:
             info['manual_install'] = True
- 
+
     # add OS builtin free alternatives to proprietary drivers
     _add_builtins(result)
- 
+
     return result
 
+
 def auto_install_filter(packages):
     '''Get packages which are appropriate for automatic installation.
 
@@ -436,6 +572,151 @@ def auto_install_filter(packages):
             result[p] = packages[p]
     return result
 
+
+class _GpgpuDriver(object):
+
+    def __init__(self, vendor=None, flavour=None):
+        self._vendors_whitelist = ('nvidia',)
+        self.vendor = vendor
+        self.flavour = flavour
+
+    def is_valid(self):
+        if self.vendor:
+            # Filter the allowed vendors
+            if not fnmatch.filter(self._vendors_whitelist, self.vendor):
+                return False
+        return not (not self.vendor and not self.flavour)
+
+
+def _process_driver_string(string):
+    '''Returns a _GpgpuDriver object'''
+    driver = _GpgpuDriver()
+    if string.find(':') != -1:
+        details = string.split(':')
+        # Remove empty strings
+        details = [x for x in details if x.strip()]
+        if len(details) != 2:
+            return None
+        for elem in details:
+            try:
+                int(elem)
+            except ValueError:
+                driver.vendor = elem
+            else:
+                driver.flavour = elem
+    else:
+        try:
+            int(string)
+        except ValueError:
+            driver.vendor = string
+        else:
+            driver.flavour = string
+
+    return driver
+
+
+def gpgpu_install_filter(packages, drivers_str):
+    drivers = []
+    allow = []
+    result = {}
+    '''Filter the Ubuntu packages according to the parameters the users passed
+
+    Ubuntu-drivers syntax
+
+    ubuntu-drivers autoinstall --gpgpu [[driver:]version]
+    ubuntu-drivers autoinstall --gpgpu driver[:version][,driver[:version]]
+
+    If no version is specified, gives the “current” supported version for the GPU in question.
+
+    Examples:
+    ubuntu-drivers autoinstall --gpgpu
+    ubuntu-drivers autoinstall --gpgpu 390
+    ubuntu-drivers autoinstall --gpgpu nvidia:390
+
+    Today this is only nvidia.  In the future there may be amdgpu-pro.
+    Possible syntax, to be confirmed only once there are driver packages that could use it:
+    ubuntu-drivers autoinstall --gpgpu nvidia:390,amdgpu
+    ubuntu-drivers autoinstall --gpgpu amdgpu:version
+    '''
+
+    if not packages:
+        return result
+
+    # No args, just --gpgpu
+    if drivers_str == 'default':
+        driver = _GpgpuDriver()
+        drivers.append(driver)
+    else:
+        # Just one driver
+        # e.g. --gpgpu 390
+        #      --gpgpu nvidia:390
+        #
+        # Or Multiple drivers
+        # e.g. --gpgpu nvidia:390,amdgpu
+        for item in drivers_str.split(','):
+            driver = _process_driver_string(item)
+            if driver and driver.is_valid():
+                drivers.append(driver)
+
+    if len(drivers) < 1:
+        return result
+
+    # If the vendor is not specified, we assume it's nvidia
+    it = 0
+    for driver in drivers:
+        if not driver.vendor:
+            drivers[it].vendor = 'nvidia'
+        it += 1
+
+    # Do not allow installing multiple versions of the nvidia driver
+    it = 0
+    vendors_temp = []
+    for driver in drivers:
+        vendor = driver.vendor
+        if vendors_temp.__contains__(vendor):
+            # TODO: raise error here
+            logging.debug('Multiple nvidia versions passed at the same time')
+            return result
+        vendors_temp.append(vendor)
+        it += 1
+
+    # If the flavour is not specified, we assume it's nvidia,
+    # and we install the newest driver
+    it = 0
+    for driver in drivers:
+        if not driver.flavour and not driver.vendor:
+            drivers[it].vendor = 'nvidia'
+        it += 1
+
+    # Filter the packages
+    # any package which matches any of those globs will be accepted
+    for driver in drivers:
+        if driver.flavour:
+            pattern = '%s*%s*' % (driver.vendor, driver.flavour)
+        else:
+            pattern = '%s*' % (driver.vendor)
+        # print('pattern: %s' % pattern)
+        allow.extend(fnmatch.filter(packages, pattern))
+        # print(allow)
+
+    # FIXME: if no flavour is specified, pick the recommended driver ?
+    # print('packages: %s' % packages)
+    for p in allow:
+        # If the version was specified, we override the recommended attribute
+        for driver in drivers:
+            if p.__contains__(driver.vendor):
+                if driver.flavour:
+                    # print('Found "%s" flavour in %s' % (driver.flavour, packages[p]))
+                    result[p] = packages[p]
+                else:
+                    # print('before recommended: %s' % packages[p])
+                    if packages[p].get('recommended'):
+                        result[p] = packages[p]
+                        # print('Found "recommended" flavour in %s' % (packages[p]))
+                break
+    return result
+
+
 def detect_plugin_packages(apt_cache=None):
     '''Get driver packages from custom detection plugins.
 
@@ -453,7 +734,7 @@ def detect_plugin_packages(apt_cache=None):
     '''
     packages = {}
     plugindir = os.environ.get('UBUNTU_DRIVERS_DETECT_DIR',
-            '/usr/share/ubuntu-drivers-common/detect/')
+                               '/usr/share/ubuntu-drivers-common/detect/')
     if not os.path.isdir(plugindir):
         logging.debug('Custom detection plugin directory %s does not exist', plugindir)
         return packages
@@ -473,7 +754,7 @@ def detect_plugin_packages(apt_cache=None):
                 exec(compile(f.read(), plugin, 'exec'), symb)
                 result = symb['detect'](apt_cache)
                 logging.debug('plugin %s return value: %s', plugin, result)
-            except Exception as e:
+            except Exception:
                 logging.exception('plugin %s failed:', plugin)
                 continue
 
@@ -492,6 +773,7 @@ def detect_plugin_packages(apt_cache=None):
 
     return packages
 
+
 def _cmp_gfx_alternatives(x, y):
     '''Compare two graphics driver names in terms of preference.
 
@@ -515,6 +797,7 @@ def _cmp_gfx_alternatives(x, y):
     assert x == y
     return 0
 
+
 def _add_builtins(drivers):
     '''Add builtin driver alternatives'''
 
@@ -536,12 +819,91 @@ def _add_builtins(drivers):
                     'free': True, 'builtin': True, 'from_distro': True, 'recommended': True}
                 break
 
+
 def get_linux_headers(apt_cache):
     '''Return the linux headers for the system's kernel'''
     kernel_detection = kerneldetection.KernelDetection(apt_cache)
     return kernel_detection.get_linux_headers_metapackage()
 
+
+def get_linux_image(apt_cache):
+    '''Return the linux image for the system's kernel'''
+    kernel_detection = kerneldetection.KernelDetection(apt_cache)
+    return kernel_detection.get_linux_image_metapackage()
+
+
+def get_linux_version(apt_cache):
+    '''Return the linux image for the system's kernel'''
+    kernel_detection = kerneldetection.KernelDetection(apt_cache)
+    return kernel_detection.get_linux_version()
+
+
 def get_linux(apt_cache):
     '''Return the linux metapackage for the system's kernel'''
     kernel_detection = kerneldetection.KernelDetection(apt_cache)
     return kernel_detection.get_linux_metapackage()
+
+
+def get_linux_modules_metapackage(apt_cache, candidate):
+    '''Return the linux-modules-$driver metapackage for the system's kernel'''
+    assert candidate is not None
+    metapackage = None
+
+    if 'nvidia' not in candidate:
+        logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping', candidate)
+        return metapackage
+
+    linux_meta = get_linux(apt_cache)
+    linux_flavour = linux_meta.replace('linux-', '')
+    candidate_flavour = candidate[candidate.rfind('-')+1:]
+
+    try:
+        int(candidate_flavour)
+    except ValueError:
+        logging.error('No flavour can be found in %s. Skipping.', candidate)
+        return metapackage
+
+    linux_modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_flavour)
+
+    try:
+        package = apt_cache.__getitem__(linux_modules_candidate)
+        # skip foreign architectures, we usually only want native
+        if (package.candidate and
+                package.candidate.architecture in ('all', system_architecture)):
+            linux_version = get_linux_version(apt_cache)
+            linux_modules_abi_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_version)
+            logging.debug('linux_modules_abi_candidate: %s' % (linux_modules_abi_candidate))
+
+            # Let's check if there is a candidate that is specific to
+            # our kernel ABI. If not, things will fail.
+            abi_specific = apt_cache.__getitem__(linux_modules_abi_candidate)
+            # skip foreign architectures, we usually only want native
+            if (abi_specific.candidate and
+                    abi_specific.candidate.architecture in ('all', system_architecture)):
+                logging.debug('Found ABI compatible %s' % (linux_modules_abi_candidate))
+                metapackage = linux_modules_candidate
+    except KeyError:
+        logging.debug('No "%s" can be found.', linux_modules_candidate)
+        pass
+
+    # Add an extra layer of paranoia, and check the availability
+    # of modules with the correct ABI
+    if metapackage:
+        return metapackage
+
+    # If no linux-modules-nvidia package is available for the current kernel
+    # we should install the relevant DKMS package
+    dkms_package = 'nvidia-dkms-%s' % candidate_flavour
+    logging.debug('Falling back to %s' % (dkms_package))
+
+    try:
+        package = apt_cache.__getitem__(dkms_package)
+        # skip foreign architectures, we usually only want native
+        if (package.candidate and
+                package.candidate.architecture in ('all', system_architecture)):
+            metapackage = dkms_package
+    except KeyError:
+        logging.error('No "%s" can be found.', dkms_package)
+        pass
+
+    return metapackage
